#!/usr/bin/python

import sys, os, random, math
sys.path.append("/home/bmmaran/workspace/py_modules/");
import cartesian, array_utils, utils

# network configuration parameters
NO_OF_NODES = 10      # no. of nodes in the network; should be >= 10
L1_FRAC = 0.1          # fraction of WAN nodes
L1_NEIGH_LOW = 1       # min # of neighbors for a WAN node
L1_NEIGH_HIGH = 3      # max # of neighbors for a WAN node
L1_LOW_RANGE = 100     # min delay between WAN nodes
L1_HIGH_RANGE = 350    # max delay between WAN nodes
L2_FRAC = 0.3          # fraction of MAN nodes
L2_LOW_RANGE = 20      # min delay between MAN nodes
L2_HIGH_RANGE = 150    # max delay between MAN nodes
L3_LOW_RANGE = 3       # min delay between LAN nodes
L3_HIGH_RANGE = 20     # max delay between LAN nodes
DIM = 2                # dimension of the topology space
COORDS_RANGE = 1000    # coordinate range of the topology space

# network parameters
delayMatrix = []
adjMatrix = []
coords = []
noOfL1Nodes = 0
noOfL2Nodes = 0
noOfL3Nodes = 0
l1Nodes = []
l2Nodes = []
l3Nodes = []

# returns the number of neighbors a node has
def countNeighbors(node, adjRow, nodeList):
    count = 0
    for i in range(len(adjRow)):
        if adjRow[i] == 1 and i != node and i in nodeList:
            count += 1
    return count

# reading network configuration parameters from config file
def readConfig():
    global NO_OF_NODES, L1_FRAC, L1_NEIGH_LOW, L1_NEIGH_HIGH
    global L1_LOW_RANGE, L1_HIGH_RANGE, L2_FRAC, L2_LOW_RANGE, L2_HIGH_RANGE 
    global L3_LOW_RANGE, L3_HIGH_RANGE, DIM, COORDS_RANGE

    if not os.access("nw_config.txt", os.R_OK):
        return
    inFile = file("nw_config.txt")
    line = inFile.readline().strip()
    while line:
        lineParts = line.split()
        if lineParts[0].find("NO_OF_NODES") != -1:
            NO_OF_NODES = int(lineParts[1])
        elif lineParts[0].find("L1_FRAC") != -1:
            L1_FRAC = float(lineParts[1])
        elif lineParts[0].find("L1_NEIGH_LOW") != -1:
            L1_NEIGH_LOW = int(lineParts[1])
        elif lineParts[0].find("L1_NEIGH_HIGH") != -1:
            L1_NEIGH_HIGH = int(lineParts[1])
        elif lineParts[0].find("L1_LOW_RANGE") != -1:
            L1_LOW_RANGE = int(lineParts[1])
        elif lineParts[0].find("L1_HIGH_RANGE") != -1:
            L1_HIGH_RANGE = int(lineParts[1])
        elif lineParts[0].find("L2_FRAC") != -1:
            L2_FRAC = float(lineParts[1])
        elif lineParts[0].find("L2_LOW_RANGE") != -1:
            L2_LOW_RANGE = int(lineParts[1])
        elif lineParts[0].find("L2_HIGH_RANGE") != -1:
            L2_HIGH_RANGE = int(lineParts[1])
        elif lineParts[0].find("L3_LOW_RANGE") != -1:
            L3_LOW_RANGE = int(lineParts[1])
        elif lineParts[0].find("L3_HIGH_RANGE") != -1:
            L3_HIGH_RANGE = int(lineParts[1])
        elif lineParts[0].find("DIM") != -1:
            DIM = int(lineParts[1])
        elif lineParts[0].find("COORDS_RANGE") != -1:
            COORDS_RANGE = int(lineParts[1])
        else:
            print "Unrecognized parameter in the network configuration file!"
        line = inFile.readline().strip()
    inFile.close()

# init program
def initNetwork(randSeed):
    global delayMatrix, adjMatrix, coords

    # seed the random number generator
    if (randSeed):
        random.seed(randSeed)
    else:
        random.seed()
    # read network configuration parameters
    readConfig()
    # initialize the delay and adjacency matrix
    delayMatrix = [[-1 for i in range(NO_OF_NODES)] for j in range(NO_OF_NODES)]
    adjMatrix  = [[0 for i in range(NO_OF_NODES)] for j in range(NO_OF_NODES)]
    for i in range(NO_OF_NODES):
        delayMatrix[i][i] = 0
        adjMatrix[i][i] = 1
    # initalize coordinates
    coords = [[COORDS_RANGE/2 for i in range(DIM)] for j in range(NO_OF_NODES)]

# choose WAN nodes
def chooseWANNodes(indexStart):
    global l1Nodes, coords, adjMatrix
    
    index = indexStart # serial number of the nodes
    l1Nodes = []
    l1Nodes.append(index)
    coords[index] = cartesian.randDistCoords([0,0], 0, COORDS_RANGE)
    while (len(l1Nodes) < noOfL1Nodes):
        centerIndex = index
        # assume to have random number of neighbors within the allowed range
        neighbors = L1_NEIGH_LOW + int((L1_NEIGH_HIGH - L1_NEIGH_LOW) * random.random())
        neighbors = min(neighbors, (noOfL1Nodes - len(l1Nodes)))
        # create neighbor nodes
        for i in range(neighbors):
            index += 1 # index of the new node
            l1Nodes.append(index)
            # new node is at a random distance from the current node
            coords[index] = cartesian.randDistCoords(coords[centerIndex], L1_LOW_RANGE, L1_HIGH_RANGE)
            adjMatrix[index][centerIndex] = adjMatrix[centerIndex][index] = 1
            # add more neighbors from the existing list
            for j in l1Nodes:
                if adjMatrix[index][j] == 0 and \
                       countNeighbors(j, adjMatrix[j], l1Nodes) < L1_NEIGH_HIGH:
                    tmpDist = cartesian.findDist(coords[j], coords[index])
                    if L1_LOW_RANGE < tmpDist and tmpDist < L1_HIGH_RANGE:
                        adjMatrix[j][index] = adjMatrix[index][j] = 1
    return index


# choose MAN nodes
def chooseMANNodes(startIndex):
    global l2Nodes, coords, adjMatrix
    
    index = startIndex
    l2Nodes = []
    while len(l2Nodes) < noOfL2Nodes:
        for l1NodeIndex in l1Nodes:
            # each WAN node has minimum one MAN child
            l2NodesPerL1Node = int(noOfL2Nodes/noOfL1Nodes)
            neighbors = 1 + int((l2NodesPerL1Node - 1)  * random.random())
            neighbors = min(neighbors, (noOfL2Nodes - len(l2Nodes)))
            for i in range(neighbors):
                index += 1
                l2Nodes.append(index)
                # assign a random coordinates with the MAN delay range
                coords[index] = cartesian.randDistCoords(coords[l1NodeIndex], L2_LOW_RANGE, L2_HIGH_RANGE)
                adjMatrix[index][l1NodeIndex] = adjMatrix[l1NodeIndex][index] = 1
                combinedNodes = l1Nodes + l2Nodes # MAN node can connect to both WAN and MAN nodes
                # check for other neighbors
                for j in combinedNodes:
                    if adjMatrix[index][j] == 0 and \
                           countNeighbors(j, adjMatrix[j], combinedNodes) < l2NodesPerL1Node:
                        tmpDist = cartesian.findDist(coords[j], coords[index])
                        if L2_LOW_RANGE < tmpDist and tmpDist < L2_HIGH_RANGE:
                            adjMatrix[j][index] = adjMatrix[index][j] = 1
    return index


# choose LAN nodes
def chooseLANNodes(startIndex):
    global l3Nodes, coords, adjMatrix

    index = startIndex
    l3Nodes = []
    while len(l3Nodes) < noOfL3Nodes:
        for l2NodeIndex in l2Nodes:
            # each MAN node has minimum one LAN child
            neighbors = 1 + int((noOfL3Nodes/noOfL2Nodes - 1)  * random.random())
            neighbors = min(neighbors, (noOfL3Nodes - len(l3Nodes)))
            for i in range(neighbors):
                index += 1
                l3Nodes.append(index)
                coords[index] = cartesian.randDistCoords(coords[l2NodeIndex], L3_LOW_RANGE, L3_HIGH_RANGE)
                # there is only MAN-LAN connection (no LAN-LAN or WAN-LAN)
                adjMatrix[l2NodeIndex][index] = adjMatrix[index][l2NodeIndex] = 1
    return index
                    

def assignNodes():
    global noOfL1Nodes, noOfL2Nodes, noOfL3Nodes
    
    # calculate the number of nodes in each level of network
    noOfL1Nodes = max(int(L1_FRAC*NO_OF_NODES), 3)
    noOfL2Nodes = int(L2_FRAC*NO_OF_NODES)
    noOfL3Nodes = NO_OF_NODES - noOfL1Nodes - noOfL2Nodes
    index = chooseWANNodes(0)
    index = chooseMANNodes(index)
    chooseLANNodes(index)

##----------- adjust the coordinates to fit into the range ---------##
def adjustCoords():
    global coords, delayMatrix, adjMatrix
    
    # calculating the max and min            
    maxCoordsVal = [None for i in range(DIM)]
    minCoordsVal = [None for i in range(DIM)]
    maxCoordsVal[0] = array_utils.findMaxCol(coords, 0)
    maxCoordsVal[1] = array_utils.findMaxCol(coords, 1)
    minCoordsVal[0] = array_utils.findMinCol(coords, 0)
    minCoordsVal[1] = array_utils.findMinCol(coords, 1)

    # shrink to fit into the coordinate range
    for i in range(DIM):
        currRange = maxCoordsVal[i] - minCoordsVal[i]
        midVal = (maxCoordsVal[i] + minCoordsVal[i])/2
        if currRange > COORDS_RANGE:
            shrinkRatio = COORDS_RANGE/currRange
            for j in range(len(coords)):
                coords[j][i] = shrinkRatio * (coords[j][i] - midVal) + midVal
            maxCoordsVal[i] = shrinkRatio * (maxCoordsVal[i] - midVal) + midVal
            minCoordsVal[i] = shrinkRatio * (minCoordsVal[i] - midVal) + midVal

    # shift the coords to the center of the coords range
    for i in range(DIM):
        midVal = (maxCoordsVal[i] + minCoordsVal[i])/2
        shift = COORDS_RANGE/2 - midVal
        for nodeCoords in coords:
            nodeCoords[i] += shift
        maxCoordsVal[i] += shift
        minCoordsVal[i] += shift

    # adjust the distances
    for i in range(NO_OF_NODES):
        for j in range(NO_OF_NODES):
            if adjMatrix[i][j] == 1:
                delayMatrix[i][j] = cartesian.findDist(coords[i], coords[j])
                delayMatrix[j][i] = delayMatrix[i][j]


##--- prepare the data file for gnuplot ---##
def createPlotData():
    # clean plot files
    fileList = os.listdir(".")
    for fileName in fileList:
        if fileName.find(".dat") != -1 and fileName.find("node") != -1:
            os.remove(fileName)
            
    for nodeIndex in l1Nodes:
        fileName = "l1_l1_node%05d.dat" % nodeIndex
        outFile = file(fileName, "w")
        for i in range(NO_OF_NODES):
            if i in l1Nodes and i != nodeIndex and adjMatrix[nodeIndex][i] == 1:
                outFile.write("%8.2f%8.2f\n" % (coords[nodeIndex][0], coords[nodeIndex][1]))
                outFile.write("%8.2f%8.2f\n" % (coords[i][0], coords[i][1]))
        outFile.close()
        fileName = "l1_l2_node%05d.dat" % nodeIndex
        outFile = file(fileName, "w")
        for i in range(NO_OF_NODES):
            if i in l2Nodes and adjMatrix[nodeIndex][i] == 1:
                outFile.write("%8.2f%8.2f\n" % (coords[nodeIndex][0], coords[nodeIndex][1]))
                outFile.write("%8.2f%8.2f\n" % (coords[i][0], coords[i][1]))
        outFile.close()

    for nodeIndex in l2Nodes:
        if countNeighbors(nodeIndex, adjMatrix[nodeIndex], l2Nodes) != 0:
            fileName = "l2_l2_node%05d.dat" % nodeIndex
            outFile = file(fileName, "w")
            for i in range(NO_OF_NODES):
                if (i in l2Nodes) and (i != nodeIndex) and (adjMatrix[nodeIndex][i] == 1):
                    outFile.write("%8.2f%8.2f\n" % (coords[nodeIndex][0], coords[nodeIndex][1]))
                    outFile.write("%8.2f%8.2f\n" % (coords[i][0], coords[i][1]))
            outFile.close()
        fileName = "l2_l3_node%05d.dat" % nodeIndex
        outFile = file(fileName, "w")
        for i in range(NO_OF_NODES):
            if (i in l3Nodes) and (adjMatrix[nodeIndex][i] == 1):
                outFile.write("%8.2f%8.2f\n" % (coords[nodeIndex][0], coords[nodeIndex][1]))
                outFile.write("%8.2f%8.2f\n" % (coords[i][0], coords[i][1]))
        outFile.close()


##--- genereate the gnuplot file ---##
def createPlotFile():
    outFile = file("plot_nw.gp", "w")
    outFile.write("set terminal postscript eps enhanced 20\n\n")
    outFile.write("set output \"nw_plot.eps\"\n\n")
    outFile.write("set xrange [0:%d]\n" % COORDS_RANGE)
    outFile.write("set yrange [0:%d]\n\n" % COORDS_RANGE)
    outFile.write("plot")
    for i in l1Nodes:
        outFile.write("\t\"l1_l1_node%05d.dat\" notitle w lp lt 1, \\\n" % i)
        outFile.write("\t\"l1_l2_node%05d.dat\" notitle w lp lt 2, \\\n" % i)
    for i in l2Nodes[:-1]:
        if countNeighbors(i, adjMatrix[i], l2Nodes) != 0:
            outFile.write("\t\"l2_l2_node%05d.dat\" notitle w lp lt 3, \\\n" % i)
        outFile.write("\t\"l2_l3_node%05d.dat\" notitle w lp lt 4, \\\n" % i)
    if countNeighbors(l2Nodes[-1], adjMatrix[l2Nodes[-1]], l2Nodes) != 0:
        outFile.write("\t\"l2_l2_node%05d.dat\" notitle w lp lt 3, \\\n" % l2Nodes[-1])
    outFile.write("\t\"l2_l3_node%05d.dat\" notitle w lp lt 4\n\n" % l2Nodes[-1])
    outFile.write("set term x11\n")
    outFile.write("set output\n")
    outFile.write("replot\n\n")
    outFile.write("pause -1\n")
    outFile.close()


##-------------- find the end to end delay matrix ---------------- ##
def calEndtoEndDelay(progBar):
    utils.shortestPathDist(delayMatrix, progBar)
    # write it to a file
    delays = []  ## not used for now
    outFile = file("delay_matrix.txt", "w")
    for row in delayMatrix:
        for val in row:
            outFile.write("%8.2f" % val)
            delays.append(val)
        outFile.write("\n")
    outFile.close()


## ----- write all the network data ------- ##
def outputNetwork():
    # write the coordinates
    outFile = file("node_coordinates.txt", "w")
    for nodeCoords in coords:
        for val in nodeCoords:
            outFile.write("%8.2f" % val)
        outFile.write("\n")
    outFile.close()
    # write the adjacency list
    outFile = file("adj_list.txt", "w")
    for i in range(len(adjMatrix)):
        for j in range(i, len(adjMatrix[i])):
            if i != j and adjMatrix[i][j] == 1:
                outFile.write("%5d%5d\n" % (i, j))
    outFile.close()
    # write the adjacency matrix
    outFile = file("adj_matrix.txt", "w")
    for row in adjMatrix:
        for val in row:
            outFile.write("%5d" % val)
        outFile.write("\n")
    outFile.close()
    # write node counts
    outFile = file("node_count.txt", "w")
    outFile.write("WAN Nodes: %d\n" % noOfL1Nodes)
    outFile.write("MAN Nodes: %d\n" % noOfL2Nodes)
    outFile.write("LAN Nodes: %d\n" % noOfL3Nodes)
    outFile.close()



## ------ the main program -------- ##
def nw_create(randSeed):
    initNetwork(randSeed)
    # create a basic random topology of WAN/MAN/LAN nodes
    assignNodes()
    print "# of WAN nodes: ", noOfL1Nodes
    print "# of MAN nodes: ", noOfL2Nodes
    print "# of LAN nodes: ", noOfL3Nodes
    # adjust the coordinates if they outside the specified range
    adjustCoords()
    # prepare gnuplot plot data files
    createPlotData()
    # prepare gnuplot plot script file
    createPlotFile()
    # calculate the all-to-all delay
    print "Calculating end-to-end delays...  ",
    calEndtoEndDelay(True)
    # write the network
    outputNetwork()
    print ""

if __name__ == '__main__':
    nw_create(None)
